"""
known_header_transformations = {

    #rsid
    'rs': 'rsid',
    'rsid': 'rsid',
    'rs_number': 'rsid',
    'rs_numbers': 'rsid',
    'snp_rsid': 'rsid',
    'rsID': 'rsid',
    'RSID': 'rsid',

    #variant_id
    'snp': 'variant_id',
    '#SNP': 'variant_id',
    'markername': 'variant_id',
    'marker': 'variant_id',
    'assay_name': 'variant_id',
    'id': 'variant_id',
    'id_dbsnp49': 'variant_id',
    'MARKER': 'variant_id',
    'snpid':'variant_id',
    'oldid':'variant_id',
    'phase1_1kg_id':'variant_id',
    'SNP': 'variant_id',
    'íd': 'variant_id',
    'MarkerName': 'variant_id',
    'MARKERNAME': 'variant_id',    
    #p_value
    'p': 'p_value',
    'pvalue': 'p_value',
    'p_value':  'p_value',
    'pval': 'p_value',
    'p_val': 'p_value',
    'gc_pvalue': 'p_value',
    'gwas_p': 'p_value',
    'frequentist_add_pvalue': 'p_value',
    'scan_p': 'p_value',
    'scanp': 'p_value',
    'p_wald': 'p_value', 
    'european_ancestry_pval_fix': 'p_value',
    'bcac_onco_icogs_gwas_p1df': 'p_value',
    'bcac_icogs1_risk_P1df': 'p_value',
    'mainp': 'p_value',
    'pv-clinical_c_k57': 'p_value',
    'frequentist_add_wald_pvalue_1': 'p_value',
    'P':'p_value',
    'P.value': 'p_value',
    'ALL.RANDOM.PVAL': 'p_value',
    'P_BOLT_LMM': 'p_value',
    'P_fathers_age_death': 'p_value',
    'P_parents_age_death': 'p_value',
    'P_top_1_percent': 'p_value',
    'P-value' : 'p_value',
    'P-val': 'p_value',
    #chromosome
    'chr': 'chromosome',
    'chromosome': 'chromosome',
    'chrom': 'chromosome',
    'scaffold': 'chromosome',
    'chr_build36': 'chromosome',
    '#chrom': 'chromosome',
    'CHR': 'chromosome',
    'Chromosome': 'chromosome',
    #base_pair_location
    'bp': 'base_pair_location',
    'pos': 'base_pair_location',
    'position': 'base_pair_location',
    'phys_pos': 'base_pair_location',
    'base_pair': 'base_pair_location',
    'basepair': 'base_pair_location',
    'base_pair_location': 'base_pair_location',
    'pos_build36': 'base_pair_location',  
    'position_b37': 'base_pair_location',
    'bp_hg19': 'base_pair_location',
    'BP': 'base_pair_location',
    'pos(b37)': 'base_pair_location',
    'POS_b37': 'base_pair_location',
    'POS' : 'base_pair_location',
    'Position': 'base_pair_location',
    'position_build36': 'base_pair_location', 
    'Position_b37': 'base_pair_location',
    'Position_hg19': 'base_pair_location',
    # chromosome combined with base pair location
    'chr_pos' : 'chr_bp',
    'chrpos' : 'chr_bp',
    'chrpos_b37' : 'chr_bp',
    'chr_pos_b37' : 'chr_bp',
    'chrpos_b36' : 'chr_bp',
    'chr_pos_b36' : 'chr_bp',
    'chrpos_b38' : 'chr_bp',
    'chr_pos_b38' : 'chr_bp',
    'chr_pos_(b36)' : 'chr_bp',
    'chr_pos_(b37)' : 'chr_bp',
    'chr_pos_(b38)' : 'chr_bp',
    'Chr': 'chr',
    #odds_ratio
    'or': 'odds_ratio',
    'odds_ratio': 'odds_ratio',
    'oddsratio': 'odds_ratio',
    'bcac_icogs1_or': 'odds_ratio',
    'OR': 'odds_ratio',
    # or range
    'L95': 'ci_upper',
    'U95': 'ci_lower',
    'orlower': 'ci_lower',
    'orupper': 'ci_upper',
    'l95': 'ci_lower',
    'u95': 'ci_upper',
    # beta
    'b': 'beta',
    'beta': 'beta',
    'effects': 'beta',
    'effect': 'beta',
    'gwas_beta': 'beta',
    'european_ancestry_beta_fix': 'beta',
    'stdbeta': 'beta',
    'bcac_onco_icogs_gwas_beta':'beta',  
    'bcac_icogs1_risk_beta': 'beta',
    'log_odds': 'beta',
    'maineffects': 'beta',
    'nbeta-clinical_c_k57': 'beta',
    'Effect': 'beta',
    'frequentist_add_beta_1:add/sle=1': 'beta',
    'ALL.RANDOM.BETA': 'beta',
    'BETA':'beta',
    'BETA_fathers_age_death': 'beta',
    'BETA_parents_age_death': 'beta',
    'BETA_top_1_percent': 'beta',
    'Beta': 'beta',
    'EFFECT': 'beta',
    'EFFECT_A1': 'beta',
    #standard_error
    'se': 'standard_error',
    'standard_error': 'standard_error',
    'stderr': 'standard_error',
    'european_ancestry_se_fix': 'standard_error',
    'bcac_onco_icogs_gwas_se': 'standard_error',
    'bcac_icogs1_risk_se': 'standard_error',
    'log_odds_se': 'standard_error',
    'mainse': 'standard_error',
    'standarderror': 'standard_error',
    'nse-clinical_c_k57': 'standard_error',
    'frequentist_add_se_1': 'standard_error',
    'StdErr': 'standard_error',
    'SE': 'standard_error',
    'SE_fathers_age_death': 'standard_error',
    'SE_parents_age_death': 'standard_error',
    'SE_top_1_percent': 'standard_error',
    'StdErr': 'standard_error',    
    'SEBETA': 'standard_error',
    'STDERR': 'standard_error',
    'se_error': 'standard_error',    
    #effect_allele
    'a1': 'effect_allele',
    'allele1': 'effect_allele',
    'allele_1': 'effect_allele',
    'effect_allele': 'effect_allele',
    'alt' : 'effect_allele',
    'inc_allele': 'effect_allele',
    'ea': 'effect_allele',
    'alleleb': 'effect_allele',
    'allele_b': 'effect_allele',
    'effectallele': 'effect_allele',
    'a1': 'effect_allele',
    'alleleB': 'effect_allele',
    'A1': 'effect_allele',
    'Allele1':'effect_allele',
    'alleleB': 'effect_allele',
    'ALLELE1': 'effect_allele',
    'EFF_ALLELE': 'effect_allele',
    'EffectAllele': 'effect_allele',
    'coded_allele': 'effect_allele',
    'Coded': 'effect_allele',
    'Effect-allele': 'effect_allele',
    #other_allele
    'a2': 'other_allele',
    'Allele2': 'other_allele',
    'allele_2': 'other_allele',
    'other_allele': 'other_allele',
    'ref': 'other_allele',
    'non_effect_allele': 'other_allele',
    'dec_allele': 'other_allele',
    'nea': 'other_allele',
    'allelea': 'other_allele',
    'allele_a': 'other_allele',
    'reference_allele': 'other_allele',
    'allele0': 'other_allele',
    'referenceallele': 'other_allele',
    'a0': 'other_allele',
    'noneffect_allele': 'other_allele',
    'alleleB': 'other_allele',
    'A2': 'other_allele',
    'alleleA': 'other_allele',
    'ALLELE0': 'other_allele',  
    'allele2': 'other_allele',
    'NONEFF_ALLELE': 'other_allele',
    'OtherAllele': 'other_allele',
    'non_coded_allele': 'other_allele',    
    'Non_coded': 'other_allele',
    'Other-allele': 'other_allele',
    #effect_allele_frequency
    'maf': 'effect_allele_frequency',
    'eafcontrols': 'effect_allele_frequency',
    'frq': 'effect_allele_frequency',
    'ref_allele_frequency': 'effect_allele_frequency',
    'frq_u': 'effect_allele_frequency',
    'f_u': 'effect_allele_frequency',
    'effect_allele_freq': 'effect_allele_frequency',
    'effect_allele_frequency': 'effect_allele_frequency',
    'freq1': 'effect_allele_frequency',
    'alt_freq': 'effect_allele_frequency',
    'a1_af': 'effect_allele_frequency',
    'bcac_onco_icogs_gwas_eaf_controls': 'effect_allele_frequency',
    'bcac_icogs1_european_controls_eaf': 'effect_allele_frequency',
    'eaf_ukb': 'effect_allele_frequency',
    'allelefreq': 'effect_allele_frequency', 
    'controls_maf': 'effect_allele_frequency',
    'effectAlleleFreq': 'effect_allele_frequency',
    'ALL.FREQ.VAR': 'effect_allele_frequency',
    'A1FREQ': 'effect_allele_frequency',   
    'freqA1': 'effect_allele_frequency',
    'EAF_UKB': 'effect_allele_frequency',
    'EAF' : 'effect_allele_frequency',
    'EFF_ALLELE_FREQ': 'effect_allele_frequency',
    'Freq': 'effect_allele_frequency',
    'FREQ_A1': 'effect_allele_frequency',
    'Coded_freq': 'effect_allele_frequency',
    'Effect-allele-frequency': 'effect_allele_frequency',
    'Freq1': 'effect_allele_frequency',
    # number of studies
    'nstudy': 'nstudy',
    'n_study': 'nstudy',
    'nstudies': 'nstudy',
    'n_studies': 'nstudy',
    # n
    'n': 'n',
    'ncase': 'n_cas',
    'cases_n': 'n_cas',
    'n_cases': 'n_cas',
    'n_controls': 'n_con',
    'n_cas': 'n_cas',
    'n_con': 'n_con',
    'n_case': 'n_cas',
    'cases_total': 'n_cas',
    'ncontrol': 'n_con',
    'controls_n': 'n_con',
    'n_control': 'n_con',
    'weight': 'n',
    'ncompletesamples': 'n',
    'controls_total': 'n_con',
    'N' : 'n',
    'N-analyzed': 'n',
    'TotalN': 'n',
    'Sample-size': 'n',
    'Sample-size-cases': 'n_cas',
    # signed statistics
    'zscore': 'z',
    'z-score': 'z',
    'gc_zscore': 'z',
    'z': 'z',
    'log_odds': 'log_odds',
    'signed_sumstat': 'signed_sumstat',
    # info
    'info': 'info',   
}
"""

header_mapper = {

    #RSID
    # ==========
    'rsid': [
        'rs',
        'rsid',
        'rs_number',
        'rs_numbers',
        'id',
        'id_dbsnp49',
        'snp_rsid',
        'rsID',
        'RSID',
        'rs_id'
        ],

    # Variant ID
    # ==========
    'variant_id': [
        'snp',
        '#SNP',
        'markername',
        'marker',
        'assay_name',
        'MARKER',
        'snpid',
        'oldid',
        'phase1_1kg_id',
        'SNP',
        'íd',
        'MarkerName',
        'MARKERNAME',
        ],

    # P-value
    # =======
    'p_value': [
        'p',
        'pvalue',
        'p_value',
        'pval',
        'p_val',
        'gc_pvalue',
        'gwas_p',
        'frequentist_add_pvalue',
        'scan_p',
        'scanp',
        'p_wald', 
        'european_ancestry_pval_fix',
        'bcac_onco_icogs_gwas_p1df',
        'bcac_icogs1_risk_P1df',
        'mainp',
        'pv-clinical_c_k57',
        'frequentist_add_wald_pvalue_1',
        'P',
        'P.value',
        'ALL.RANDOM.PVAL',
        'P_BOLT_LMM',
        'P_fathers_age_death',
        'P_parents_age_death',
        'P_top_1_percent',
        'P-value',
        'P-val'
        ],

    # Chromosome
    # ==========
    'chromosome': [
        'chr',
        'chromosome',
        'chrom',
        'scaffold',
        'chr_build36',
        '#chrom',
        'CHR',
        'Chromosome'
        ],

    # Base pair location
    # ==================
    'base_pair_location': [
        'bp',
        'pos',
        'position',
        'phys_pos',
        'base_pair',
        'basepair',
        'base_pair_location',
        'pos_build36',  
        'position_b37',
        'bp_hg19',
        'BP',
        'pos(b37)',
        'POS_b37',
        'POS' ,
        'Position',
        'position_build36', 
        'Position_b37',
        'Position_hg19',
        'Pos_GRCh37',
        'GENPOS'
        ],
    # Chromosome and base pair location
    # ================================
    'chromosome_bp': [
    'chr_bp',
    'chr_pos',
    'chrpos',
    'chrpos_b37',
    'chr_pos_b37',
    'chrpos_b36',
    'chr_pos_b36',
    'chrpos_b38',
    'chr_pos_b38',
    'chr_pos_(b36)',
    'chr_pos_(b37)',
    'chr_pos_(b38)',
    ],

    # Odds ratio
    # ==========
    'odds_ratio': [
        'or',
        'odds_ratio',
        'oddsratio',
        'bcac_icogs1_or',
        'OR'
        ],

    # Upper confidence interval
    # =========================
    'ci_upper': [
        'U95',
        'orupper',
        'u95'
        ],

    # Lower confidence interval
    # =========================
    'ci_lower': [
        'L95',
        'orlower',
        'l95'
        ],

    # Beta
    # ====
    'beta': [
        'b',
        'beta',
        'effects',
        'effect',
        'gwas_beta',
        'european_ancestry_beta_fix',
        'stdbeta',
        'bcac_onco_icogs_gwas_beta',  
        'bcac_icogs1_risk_beta',
        'log_odds',
        'maineffects',
        'nbeta-clinical_c_k57',
        'Effect',
        'frequentist_add_beta_1:add/sle=1',
        'ALL.RANDOM.BETA',
        'BETA',
        'BETA_fathers_age_death',
        'BETA_parents_age_death',
        'BETA_top_1_percent',
        'Beta',
        'EFFECT',
        'EFFECT_A1'
        ],

    # Standard error
    # ==============
    'standard_error': [
        'se',
        'standard_error',
        'stderr',
        'european_ancestry_se_fix',
        'bcac_onco_icogs_gwas_se',
        'bcac_icogs1_risk_se',
        'log_odds_se',
        'mainse',
        'standarderror',
        'nse-clinical_c_k57',
        'frequentist_add_se_1',
        'StdErr',
        'SE',
        'SE_fathers_age_death',
        'SE_parents_age_death',
        'SE_top_1_percent',
        'StdErr',    
        'SEBETA',
        'STDERR',
        'se_error'
        ],

    # Effect allele
    # =============
    'effect_allele': [
        'a1',
        'allele1',
        'allele_1',
        'effect_allele',
        'alt' ,
        'inc_allele',
        'ea',
        'alleleb',
        'allele_b',
        'effectallele',
        'a1',
        'alleleB',
        'A1',
        'Allele1',
        'alleleB',
        'ALLELE1',
        'EFF_ALLELE',
        'EffectAllele',
        'coded_allele',
        'Coded',
        'Effect-allele'
        ],
    
    # Other allele
    # ============
    'other_allele': [
        'a2',
        'Allele2',
        'allele_2',
        'other_allele',
        'ref',
        'non_effect_allele',
        'dec_allele',
        'nea',
        'allelea',
        'allele_a',
        'reference_allele',
        'allele0',
        'referenceallele',
        'a0',
        'noneffect_allele',
        'alleleB',
        'A2',
        'alleleA',
        'ALLELE0',  
        'allele2',
        'NONEFF_ALLELE',
        'OtherAllele',
        'non_coded_allele',    
        'Non_coded',
        'Other-allele'
        ],

    # Effect allele frequency
    # =======================
    'effect_allele_frequency': [
        'maf',
        'eafcontrols',
        'frq',
        'ref_allele_frequency',
        'frq_u',
        'f_u',
        'effect_allele_freq',
        'effect_allele_frequency',
        'freq1',
        'freq',
        'alt_freq',
        'a1_af',
        'bcac_onco_icogs_gwas_eaf_controls',
        'bcac_icogs1_european_controls_eaf',
        'eaf_ukb',
        'allelefreq', 
        'controls_maf',
        'effectAlleleFreq',
        'ALL.FREQ.VAR',
        'A1FREQ',   
        'freqA1',
        'EAF_UKB',
        'EAF' ,
        'EFF_ALLELE_FREQ',
        'Freq',
        'FREQ_A1',
        'Coded_freq',
        'Effect-allele-frequency',
        'Freq1',
        'A1FREQ'
        ],
    
    # Number of studies
    # =================
    'nstudy': [
        'nstudy',
        'n_study',
        'nstudies',
        'n_studies'
        ],

    # n
    # =
    'n': [
        'n',
        'N' ,
        'N-analyzed',
        'TotalN',
        'Sample-size',
        'weight',
        'ncompletesamples'
        ],

    # n cases
    # =======
    'n_cas': [
        'ncase',
        'cases_n',
        'n_cases',
        'n_cas',
        'n_case',
        'cases_total',
        'Sample-size-cases'
        ],

    # n cases
    # =======
    'n_con': [
        'ncontrol',
        'controls_n',
        'n_control',
        'n_con',
        'n_controls',
        'controls_total'
        ],

    # signed statistics
    # =================
    'z': [
        'zscore',
        'z-score',
        'gc_zscore',
        'z'
        ],
    }



from pydantic import BaseModel, Field
from typing import List, Optional
import json

class FileConfig (BaseModel):
    outFileSuffix: str = Field(default="formatted_")
    fieldSeparator: str = Field(default="tab")
    removeComments: str = Field(default=None)
    naValue: str = Field(default=None)
    convertNegLog10Pvalue: bool = Field(default=False)

class SplitConfig (BaseModel):
    field: Optional[str]=None
    separator: Optional[str]=None
    capture: Optional[str]=None
    new_field:Optional[List]=None
    include_original:Optional[bool]=False

class EditConfig (BaseModel):
    field: Optional[str]=None
    rename: Optional[str]=None
    find: Optional[str]=None
    replace: Optional[str]=None
    extract: Optional[str]=None

class ColumnConfig (BaseModel):
    split: List[SplitConfig]
    edit: List[EditConfig]

class Formatconfig (BaseModel):
    fileConfig: FileConfig
    columnConfig: List[ColumnConfig]

"""
#Example to show the schema structure of the json file
file_config=FileConfig()

column_config_1 = ColumnConfig(
    field="gene",
    split=SplitConfig(),
    edit=EditConfig()
)
column_config_2 = ColumnConfig(
    field="chromsome",
    split=SplitConfig(),
    edit=EditConfig()
)
format_config = Formatconfig(fileConfig=file_config,columnConfig=[column_config_1,column_config_2])

print(json.loads(format_config.json()))
example: {'fileConfig': {'outFilePrefix': 'formatted_', 'md5': False, 'convertNegLog10Pvalue': False, 'fieldSeparator': 'tab', 'removeComments': None}, 'columnConfig': [{'field': 'gene', 'split': {'leftName': None, 'rightName': None, 'separator': None}, 'edit': {'rename': None, 'find': None, 'replace': None, 'extract': None}}, {'field': 'chromsome', 'split': {'leftName': None, 'rightName': None, 'separator': None}, 'edit': {'rename': None, 'find': None, 'replace': None, 'extract': None}}]}
"""


#----------------REGENIE----------------
regenie = {
    "fileConfig": {
        "outFileSuffix": "regenie_formatted",
        "fieldSeparator": " ",
        "naValue": "NA",
        "convertNegLog10Pvalue": True,
        "removeComments": None
    },
    "columnConfig": {
        "split": [
        ],
        "edit": [
            {
                "field": "CHROM",
                "rename": "chromosome",
                "find": None,
                "replace": None,
                "extract": None
            },
            {
                "field": "GENPOS",
                "rename": "base_pair_location",
                "find": None,
                "replace": None,
                "extract": None
            },
            {
                "field": "ID",
                "rename": "ID",
                "find": None,
                "replace": None,
                "extract": None
            },
            {
                "field": "ALLELE0",
                "rename": "other_allele",
                "find": None,
                "replace": None,
                "extract": None
            },
            {
                "field": "ALLELE1",
                "rename": "effect_allele",
                "find": None,
                "replace": None,
                "extract": None
            },
            {
                "field": "A1FREQ",
                "rename": "effect_allele_frequency",
                "find": None,
                "replace": None,
                "extract": None
            },
            {
                "field": "N",
                "rename": "n",
                "find": None,
                "replace": None,
                "extract": None
            },
            {
                "field": "BETA",
                "rename": "beta",
                "find": None,
                "replace": None,
                "extract": None
            },
            {
                "field": "SE",
                "rename": "standard_error",
                "find": None,
                "replace": None,
                "extract": None
            },
            {
                "field": "LOG10P",
                "rename": "p_value",
                "find": None,
                "replace": None,
                "extract": None
            }
        ]
    }
}

#----------------BOLT-LMM----------------
boltlmm = {
    "fileConfig": {
        "outFileSuffix": None,
        "fieldSeparator": None,
        "naValue": None,
        "convertNegLog10Pvalue": False,
        "removeComments": False
    },
    "columnConfig": {
        "split": [
        ],
        "edit": [
            {
                "field": "SNP",
                "rename": "variant_id",
                "find": None,
                "replace": None,
                "extract": None
            },
            {
                "field": "CHR",
                "rename": "chromosome",
                "find": None,
                "replace": None,
                "extract": None
            },
            {
                "field": "BP",
                "rename": "base_pair_location",
                "find": None,
                "replace": None,
                "extract": None
            },
            {
                "field": "GENPOS",
                "rename": "GENPOS",
                "find": None,
                "replace": None,
                "extract": None
            },
            {
                "field": "ALLELE1",
                "rename": "effect_allele",
                "find": None,
                "replace": None,
                "extract": None
            },
            {
                "field": "ALLELE0",
                "rename": "other_allele",
                "find": None,
                "replace": None,
                "extract": None
            },
            {
                "field": "A1FREQ",
                "rename": "effect_allele_frequency",
                "find": None,
                "replace": None,
                "extract": None
            },
            {
                "field": "INFO",
                "rename": "INFO",
                "find": None,
                "replace": None,
                "extract": None
            },
            {
                "field": "BETA",
                "rename": "beta",
                "find": None,
                "replace": None,
                "extract": None
            },
            {
                "field": "SE",
                "rename": "standard_error",
                "find": None,
                "replace": None,
                "extract": None
            },
            {
                "field": "P_BOLT_LMM_INF",
                "rename": "P_BOLT_LMM_INF",
                "find": None,
                "replace": None,
                "extract": None
            },
            {
                "field": "P_BOLT_LMM",
                "rename": "p_value",
                "find": None,
                "replace": None,
                "extract": None
            }
        ]
    }
}

#----------------SNPTEST----------------
snptest = {
    "fileConfig": {
        "outFileSuffix": None,
        "fieldSeparator": " ",
        "naValue": None,
        "convertNegLog10Pvalue": False,
        "removeComments": "#"
    },
    "columnConfig": {
        "split": [
        ],
        "edit": [
            {
                "field": "alternate_ids",
                "rename": "variant_id",
                "find": None,
                "replace": None,
                "extract": None
            },
            {
                "field": "rsid",
                "rename": "rsid",
                "find": None,
                "replace": None,
                "extract": None
            },
            {
                "field": "chromosome",
                "rename": "chromosome",
                "find": None,
                "replace": None,
                "extract": None
            },
            {
                "field": "position",
                "rename": "base_pair_location",
                "find": None,
                "replace": None,
                "extract": None
            },
            {
                "field": "alleleA",
                "rename": "other_allele",
                "find": None,
                "replace": None,
                "extract": None
            },
            {
                "field": "alleleB",
                "rename": "effect_allele",
                "find": None,
                "replace": None,
                "extract": None
            },
            {
                "field": "all_total",
                "rename": "n",
                "find": None,
                "replace": None,
                "extract": None
            },
            {
                "field": "frequentist_add_pvalue",
                "rename": "p_value",
                "find": None,
                "replace": None,
                "extract": None
            },
            {
                "field": "frequentist_add_beta_1",
                "rename": "beta",
                "find": None,
                "replace": None,
                "extract": None
            },
            {
                "field": "frequentist_add_se_1",
                "rename": "standard_error",
                "find": None,
                "replace": None,
                "extract": None
            }
        ]
    }
}

#----------------SAIGE------------------
saige={
    "fileConfig": {
        "outFileSuffix": None,
        "fieldSeparator": "\t",
        "naValue": None,
        "convertNegLog10Pvalue": False,
        "removeComments": None
    },
    "columnConfig": {
        "split": [
        ],
        "edit": [
            {
                "field": "CHR",
                "rename": "chromosome",
                "find": None,
                "replace": None,
                "extract": None
            },
            {
                "field": "POS",
                "rename": "base_pair_location",
                "find": None,
                "replace": None,
                "extract": None
            },
            {
                "field": "MarkerID",
                "rename": "variant_id",
                "find": None,
                "replace": None,
                "extract": None
            },
            {
                "field": "Allele1",
                "rename": "other_allele",
                "find": None,
                "replace": None,
                "extract": None
            },
            {
                "field": "Allele2",
                "rename": "effect_allele",
                "find": None,
                "replace": None,
                "extract": None
            },
            {
                "field": "AC_Allele2",
                "rename": "effect_allele_count",
                "find": None,
                "replace": None,
                "extract": None
            },
            {
                "field": "AF_Allele2",
                "rename": "effect_allele_frequency",
                "find": None,
                "replace": None,
                "extract": None
            },
            {
                "field": "MissingRate",
                "rename": "MissingRate",
                "find": None,
                "replace": None,
                "extract": None
            },
            {
                "field": "BETA",
                "rename": "beta",
                "find": None,
                "replace": None,
                "extract": None
            },
            {
                "field": "SE",
                "rename": "standard_error",
                "find": None,
                "replace": None,
                "extract": None
            },
            {
                "field": "Tstat",
                "rename": "Tstat",
                "find": None,
                "replace": None,
                "extract": None
            },
            {
                "field": "var",
                "rename": "var",
                "find": None,
                "replace": None,
                "extract": None
            },
            {
                "field": "p.value",
                "rename": "p_value",
                "find": None,
                "replace": None,
                "extract": None
            },
            {
                "field": "N",
                "rename": "n",
                "find": None,
                "replace": None,
                "extract": None
            }
        ]
    }
}
#----------------pre_defined_configure----------------
pre_defined_configure={
    "REGENIE": regenie,
    "BOLT-LMM": boltlmm,
    "SNPTEST": snptest,
    "SAIGE": saige
}


from pathlib import Path
from typing import Union
import pandas as pd
from pandas.io.parsers import TextFileReader
import numpy as np
import petl as etl
import pandas as pd


"""formatters

rename_headers(chromosome=...,)

order_headers(custome_order=False)

replace_values(all=False, field, from, to)
    petl.transform.conversions.replaceall()
    petl.transform.conversions.replace()

coerce_chromosome()

normalise_missing_values()

"""


class SumStatsTable:
    FIELD_MAP = {"variant_id": "rsid"}
    FIELDS_REQUIRED = ("chromosome", "base_pair_location", "effect_allele",
                       "other_allele", "standard_error",
                       "effect_allele_frequency", "p_value")
    FIELDS_EFFECT = ("beta", "odds_ratio", "hazard_ratio")
    FIELDS_OPTIONAL = ("variant_id", "rsid", "info", "ci_upper", "ci_lower", "ref_allele")

    def __init__(self, sumstats_file: Path, delimiter: str = None, removecomments: str = None) -> None:
        self.filename = str(sumstats_file)
        self.delimiter = delimiter if delimiter else self._get_delimiter(sumstats_file)
        self.removecomments = removecomments if removecomments else None
        self.sumstats = self.from_file()
        

    def reformat_header(self, header_map: dict = FIELD_MAP) -> etl.Table:
        """Reformats the headers according to the standard

        Returns:
            etl.Table
        """
        self.rename_headers(header_map=header_map)
        missing_headers = self._get_missing_headers()
        if missing_headers:
            self._add_missing_headers(missing_headers)
        header_order = self._set_header_order()
        self.sumstats = etl.cut(self.sumstats, *header_order)
        return self.sumstats
    
    def map_header(self) -> etl.Table:
        """Reformats the headers according to the standard

        Returns:
            etl.Table
        """
        missing_headers = self._get_missing_headers()
        if missing_headers:
            self._add_missing_headers(missing_headers)
        header_order = self._set_header_order()
        self.sumstats = etl.cut(self.sumstats, *header_order)
        return self

    def from_file(self) -> Union[etl.Table, None]:
        """Try to read the file in to a Table.
        Files can be TAB seperated and optionally compressed
        with (B)GZIP. There could be cases where an input file 
        has been renamed but the data is something different 
        to that suggested by the name and extension. Most 
        cases should be covered by the exception clause.

        Arguments:
            infile -- Input file

        Returns:
            petl Table or None
        """
        try:
            if len(self.delimiter) == 1:
                self.sumstats = etl.fromcsv(self.filename, delimiter=self.delimiter,skipinitialspace=True)
                if self.removecomments is not None:
                    self.sumstats = etl.skipcomments(self.sumstats,self.removecomments)
            else:
                df = pd.read_csv(self.filename ,sep=self.delimiter, comment=self.removecomments)
                self.sumstats = etl.fromdataframe(df)
            
            if not self.is_table_content():
                return None
            return self.sumstats
        except (IOError, UnicodeDecodeError) as exception:
            print(exception)
            return None

    def is_table_content(self) -> bool:
        """Bool for whether table content exists

        Returns:
            Boolean
        """
        return etl.nrows(self.head_table(nrows=1)) > 0
    
    def to_file(self, outfile: Path) -> None:
        """Write table to TSV file

        Arguments:
            outfile -- Output file name
        """
        self.sumstats.totsv(str(outfile))

    def head_table(self, nrows: int = 10) -> etl.Table:
        return etl.head(self.sumstats, n=nrows)
    
    def example_table(self, nrows: int = 5) -> etl.Table:
        self.sumstats = etl.head(self.sumstats, n=nrows)
        return self

    def _get_delimiter(self, filepath: Path) -> str:
        """Get delimiter from file path

        Arguments:
            filepath -- Input file path

        Returns:
            delimiter, either ',' or '\t'
        """
        if not isinstance(filepath, Path):
            filepath = Path(filepath)
        if '.csv' in filepath.suffixes:
            return ','
        elif '.txt' in filepath.suffixes:
            return ' '
        else:
            return '\t'

    def rename_headers(self, header_map: dict) -> etl.Table:
        """Rename headers according to the header map

        Keyword Arguments:
            header_map -- dict of header mappings from old to new \
                (default: {HEADER_MAP})

        Returns:
            etl.Table
        """
        filtered_header_map = {k: v for k, v in header_map.items() if k in self.header()}
        self.sumstats = etl.rename(self.sumstats, filtered_header_map)
        return self

    def normalise_missing_values(self,na_value:str) -> etl.Table:
        self.sumstats = etl.replaceall(self.sumstats, 'NA', '#NA')
        self.sumstats = etl.replaceall(self.sumstats, None, '#NA')
        self.sumstats = etl.replaceall(self.sumstats, '', '#NA')
        
        if na_value is not None:
            self.sumstats = etl.replaceall(self.sumstats, na_value, '#NA')   
        return self
    
    def convert_neg_log10_pvalue(self) -> etl.Table:
        self.sumstats = etl.convert(self.sumstats, 'p_value', lambda x: 10**(-float(x)))
        return self

    def _get_missing_headers(self) -> set:
        """Identify and return missing headers

        Returns:
            set of missing headers
        """
        missing_headers = set(self.FIELDS_REQUIRED) - set(self.header())
        if set(self.FIELDS_EFFECT).isdisjoint(set(self.header())):
            missing_headers.add("beta")
        return missing_headers

    def _set_header_order(self) -> list:
        """Set the header order

        Returns:
            List of headers in standard order
        """
        all_headers = [h for h in self.FIELDS_REQUIRED]
        all_headers.extend([h for h in self.FIELDS_OPTIONAL])
        all_headers.extend([h for h in self.FIELDS_EFFECT])
        header_order = [h for h in self.FIELDS_REQUIRED]
        header_order.extend([h for h in self.FIELDS_OPTIONAL if h in self.header()])
        header_order.extend([h for h in self.header() if h not in all_headers])
        if 'beta' in self.header():
            header_order.insert(4, 'beta')
            for h in ('odds_ratio', 'hazard_ratio'):
                header_order.append(h) if h in self.header() else None
        elif 'beta' not in self.header() and 'odds_ratio' in self.header():
            header_order.insert(4, 'odds_ratio')
            header_order.append('hazard_ratio') if 'hazard_ratio' in self.header() else None
        elif 'odds_ratio' not in self.header() and 'hazard_ratio' in self.header():
            header_order.insert(4, 'hazard_ratio')
        else:
            pass
        return header_order

    def _add_missing_headers(self, missing_headers: set) -> etl.Table:
        """Add missing headers and fill with #NA

        Arguments:
            missing_headers -- missing headers set

        Returns:
            etl.Table
        """
        add_fields = [(h, '#NA') for h in missing_headers]
        if len(add_fields) > 1:
            self.sumstats = etl.addfields(self.sumstats, add_fields)
        else:
            field, value = add_fields[0]
            self.sumstats = etl.addfield(self.sumstats, field, value)
        return self.sumstats

    def header(self) -> tuple:
        """Get the header of the file

        Returns:
            tuple of the headers
        """
        if self.is_table_content():
            return etl.header(self.sumstats)
        return ()

    def effect_field(self) -> Union[str, None]:
        """Get the effect allele (field index 4).

        Returns:
            effect field label
        """
        field_4 = self._get_field_label_from_index(4)
        return field_4

    def p_value_field(self) -> Union[str, None]:
        """Get the p_value field (field index 7).

        Returns:
            p_value field label
        """
        field_4 = self._get_field_label_from_index(7)
        return field_4

    def _get_field_label_from_index(self, index: int) -> Union[str, None]:
        """Get the field label from specified index

        Arguments:
            index -- index of field

        Returns:
            field label or None
        """
        return self.header()[index] if len(self.header()) > index else None

    def _pval_to_mantissa_and_exponent(self, table: etl.Table) -> etl.Table:
        table_w_split_p = etl.split(self._square_up_table(table),
                                    self.p_value_field(),
                                    'e|E',
                                    newfields=['_p_value_mantissa',
                                               '_p_value_exponent'],
                                    include_original=True,
                                    maxsplit=1)
        return table_w_split_p

    def pval_to_mantissa_and_exponent(self, df: pd.DataFrame) -> pd.DataFrame:
        """Create the mantissa and exponent columns from the pvalue field.
        First a row needs to be appended (we use index -99) with the delimiter,
        else the split cannot be done if no delimiters were found. The row is
        removed by index before returning the dataframe.

        Arguments:
            df -- dataframe

        Returns:
            dataframe
        """
        mant_and_exp = ['_p_value_mantissa', '_p_value_exponent']
        psplit_row = pd.Series({self.p_value_field(): '1000e1000'}, name=-99)
        df = pd.concat([df, psplit_row.to_frame().T], ignore_index=False)
        df[mant_and_exp] = (df[self.p_value_field()]
                            .str.split(r"e|E",
                                       regex=True,
                                       n=1,
                                       expand=True))
        df = df.drop(index=-99)
        return df

    def _prep_table_for_validation(self) -> etl.Table:
        table = etl.Table()
        if self.is_table_content():
            table = self._pval_to_mantissa_and_exponent(table=self.sumstats)
        return table

    def as_pd_df(self,
                 nrows: int = None,
                 chunksize: int = None,
                 skiprows: int = None) -> Union[pd.DataFrame, TextFileReader]:
        """Sumstats table as a Pandas dataframe or dataframe
        iterator (TextFileReader)

        Keyword Arguments:
            nrows -- Number of rows (default: {None, which means all rows})
            chunksize -- Number of rows to store in mem at once
            skiprows -- Number of rows to skip from start of file

        Returns:
            Pandas dataframe or iter
        """
        df = pd.DataFrame()
        skip = range(1, skiprows) if skiprows else None
        if self.is_table_content():
            df = pd.read_table(self.filename,
                               sep=self.delimiter,
                               chunksize=chunksize,
                               nrows=nrows,
                               na_values=["", "#NA", "NA", "N/A", "NaN", "NR"],
                               dtype=str,
                               skiprows=skip
                               )
        return df

    def _square_up_table(self, table: etl.Table, missing: str = "#NA") -> etl.Table:
        """Square up a table with missing/extra values on rows.

        Returns:
            etl.Table
        """
        return etl.cat(table, missing=missing)
    
    def split_columns_by_separator(self, field, separator: str, newfields: list,include_original:bool) -> etl.Table:
        """
        split(table, field, pattern, newfields=None, include_original=False, maxsplit=0, flags=0)
        """
        self.sumstats = etl.split(self.sumstats,field,separator,newfields,include_original=include_original)
        return self
    
    def split_capture(self, field, pattern, newfields: list,include_original:bool):
        """
        petl.sub(table, field, pattern, repl, count=0, flags=0)
        OR 
        petl.capture(table, field, pattern, newfields=None, include_original=False, flags=0)
        Convenience function to convert values under the given field using a regular expression substitution. See also re.sub().
        """
        self.sumstats = etl.capture(self.sumstats, field, pattern, newfields, include_original=include_original)
        return self
    
    def find_and_replace(self, field, find, replace) -> etl.Table:
        """
        petl.replace(table, field, a, b)
        Only applied to the colunns but not the header

        petl.sub(table, field, pattern, repl, count=0, flags=0)
        pattern can be str or regex
        """
        self.sumstats = etl.sub(self.sumstats, field, find, replace, count=0, flags=0)
        return self
    
    def extract(self, field, pattern, newfield):
        """
        capture function to extract regex pattern to original column
        """
        regexs=f"({pattern})"
        newfields=[newfield]
        self.sumstats = etl.capture(self.sumstats, field, regexs, newfields, include_original=False)
        return self

    def _covert_value (self, table: etl.Table, field, value, replace) -> etl.Table:
        """
        petl.transform.conversions.convert()
        """
        return etl.convert(table,field,'replace',value,replace)
    
    



from pathlib import Path
from typing import Union
import yaml
import petl as etl

from gwas_sumstats_tools.interfaces.data_table import SumStatsTable
from gwas_sumstats_tools.interfaces.metadata import (MetadataClient,
                                                     SumStatsMetadata)
from gwas_sumstats_tools.utils import exit_if_no_data


class Reader():
    """Class for reading summary statistics data tables
    and associated metadata
    """
    def __init__(self,
                 sumstats_file: Path = None,
                 delimiter: str = None,
                 remove_comments: str = None,
                 metadata_file: Path = None) -> None:
        if not metadata_file and isinstance(sumstats_file, Path):
            metadata_file = sumstats_file.with_suffix(sumstats_file.suffix + "-meta.yaml")

        self.delimiter=delimiter.encode().decode('unicode_escape') if delimiter else None
        
        self.removecomments = remove_comments if remove_comments else None

        self.data = SumStatsTable(sumstats_file=sumstats_file, delimiter=self.delimiter, removecomments=self.removecomments) if sumstats_file else None
        
        self.meta = MetadataClient(in_file=metadata_file) if metadata_file else None

    def file_header(self) -> Union[SumStatsTable.header, None]:
        """Get the file header

        Returns:
            Sumstats data file header
        """
        if self.data:
            return self.data.header()
        else:
            return None

    def metadata_model(self, **kwargs) -> Union[SumStatsMetadata, None]:
        """Get the metadata model

        Keyword Arguments:
            **kwargs to pydantic model.dict

        Returns:
            Metadata model
        """
        if self.meta:
            self.meta.from_file()
            return self.meta.metadata
        else:
            return None

    def metadata_dict(self, include: list = None, **kwargs) -> dict:
        """Get the metadata as a dict

        Keyword Arguments:
            include -- Optional list of fields to include (default: include all)
            **kwargs to pydantic model.dict()

        Returns:
            Dict of metadata
        """
        metadata_dict = {}
        if self.meta:
            self.meta.from_file()
            if include:
                metadata_dict = {field: self.meta.metadata.dict(**kwargs).get(field)
                                 for field in include}
            else:
                return self.meta.metadata.dict(**kwargs)
        return metadata_dict

    def head(self, **kwargs) -> Union[etl.Table, None]:
        """Head of data file

        Returns:
            etl.Table
        """
        if self.data:
            return self.data.head_table(**kwargs)
        else:
            return None


def read(filename: Path,
         metadata_infile: Path = None,
         get_header: bool = False,
         get_all_metadata: bool = False,
         get_metadata: list = None,
         remove_comments: str = None,
         delimiter: str = None):
    """Driver function for the Reader class

    Arguments:
        filename -- sumstats filename

    Keyword Arguments:
        metadata_infile -- metadata filename (default: {None})
        get_header -- return the header (default: {False})
        get_all_metadata -- return all the metadata  (default: {False})
        get_metadata -- return the metadata fields specified in this list (default: {None})

    Returns:
        _description_
    """
    reader = Reader(sumstats_file=filename,
                    delimiter=delimiter,
                    remove_comments=remove_comments,
                    metadata_file=metadata_infile)
    exit_if_no_data(reader.data.sumstats)
    if get_header:
        message = "[bold]\n#-------- SUMSTATS HEADERS --------#\n[/bold]"
        return (reader.file_header(), message)
    if get_all_metadata:
        message = "[bold]\n#-------- SUMSTATS METADATA --------#\n[/bold]"
        return (yaml.dump(reader.metadata_dict()), message)
    if get_metadata:
        message = "[bold]\n#-------- SUMSTATS METADATA --------#\n[/bold]"
        return (yaml.dump(reader.metadata_dict(include=get_metadata)), message)
    if not any([get_header, get_all_metadata, get_metadata]):
        message = "[bold]\n#-------- SUMSTATS DATA PREVIEW --------#\n[/bold]"
        return (reader.head(), message)



"""
A pydantic (https://docs.pydantic.dev/) schema that can be 
used by python projects for defining the metadata model.
Keep the same scheme with https://github.com/EBISPOT/gwas-summary-statistics-standard/blob/master/schema/metadata-yamale-schema.yaml
"""

from pydantic import (BaseModel,
                      constr)
from datetime import date
from typing import List, Optional
from enum import Enum
"""
Enums
"""


class SexEnum(str, Enum):
    male = 'M'
    female = 'F'
    combined = 'combined'
    unknown='NR'


class CoordinateSystemEnum(str, Enum):
    zero = '0-based'
    one = '1-based'


"""
Models
"""


class SampleMetadata(BaseModel):
    sample_ancestry_category: Optional[List[str]] = None
    sample_ancestry: Optional[List[str]] = None
    sample_size: Optional[int] = None
    ancestry_method: Optional[List[str]] = None
    case_control_study: Optional[bool] = None
    case_count: Optional[int] = None
    control_count: Optional[int] = None


class SumStatsMetadata(BaseModel):
    # Study meta-data
    gwas_id: Optional[constr(regex=r'^GCST\d+$')] = None
    author_notes: Optional[str] = None
    gwas_catalog_api: Optional[str] = None
    date_metadata_last_modified: date
    # Trait Information
    trait_description: Optional[List[str]] = None
    ontology_mapping: Optional[List[str]] = None
    # Genotyping Information
    genome_assembly: str
    coordinate_system: Optional[CoordinateSystemEnum] = None
    genotyping_technology: Optional[List[str]] = None
    imputation_panel: Optional[str] = None
    imputation_software: Optional[str] = None
    # Sample Information
    samples: Optional[List[SampleMetadata]] = None
    sex: Optional[SexEnum] = None
    # Summary Statistic information:
    data_file_name: str
    file_type: str
    data_file_md5sum: str
    analysis_software: Optional[str] = None
    adjusted_covariates: Optional[List[str]] = None
    minor_allele_freq_lower_limit: Optional[float] = None
    # Harmonization status
    is_harmonised: Optional[bool] = None
    is_sorted: Optional[bool] = None
    harmonisation_reference: Optional[str] = None
    
    class Config:
        title = 'GWAS Summary Statistics metadata schema'
        use_enum_values = True
